#include <EEPROM.h>
#include <WiFi.h>
#include <WebServer.h>
#include <ESPmDNS.h>
#include <ESP32Servo.h>

// define variables and structs

String version = "ESafeP 1.3";  // 1.0: Web Interface finished; 1.1: Servo motor works now; 1.2: The 3x4 Keyboard functions; 1.3: Added Keyboard functionality: Change the Code

int currentWiFiAP = 1;  // 1: normal WiFi; 2: alternative WiFi; 3: softAP (ESP opens an access point)
char apSSID[10] = "ESafeP-AP";
char apPSK[14] = "aA161616161Aa";
bool apON = false;

struct WiFiConfig {
  char ssid[33];
  char pwd[64];
  char alternative_ssid[33];
  char alternative_pwd[64];
};
struct Config {
  char code[9];
  char friendly_name[21];
};
int wiFiConfigAddress = 4;  // this is a pointer where objects are stored on EEPROM (points to the end; points to free space); at 0 the state of establishment is stored (0 = not established or 1 established)
int configAddress = wiFiConfigAddress + sizeof(WiFiConfig);

int doorServoPin = 12;  // pin number where the control line of the servo motor is connected to
Servo doorServo;  // servo object for the door
bool safeOpened = false;  // value if the safe is opened or closed
int servo_open_angle = 90;
int servo_closed_angle = 135;

// All pins for I/O (Keyboard, LEDs, etc.)
int keyboard3x4_column1 = 25;
int keyboard3x4_column2 = 26;
int keyboard3x4_column3 = 27;
int keyboard3x4_row1 = 33;
int keyboard3x4_row2 = 32;
int keyboard3x4_row3 = 35;
int keyboard3x4_row4 = 34;
int readyForInputLED = 2;
int wrongCodeLED = 4;
int codeChangeLED = 5;
int newCodeLED = 18;
int newCodeRepeatLED = 19;
int invalidCodeLED = 21;
int typeOldCodeLED = 22;


int webport = 80; // the port for the web interface
WebServer server(webport);  // the webserver (port "webport")

void setup() {
  pinSetup();  // set up all needed pins

  doorServo.attach(doorServoPin);
  
  Serial.begin(115200);
  EEPROM.begin(4 + sizeof(WiFiConfig) + sizeof(Config));  // load the data on EEPROM into the RAM
  connectWiFi();  // connect to wifi on startup
  webSetup();  // register the web paths and start the web server
  if (MDNS.begin("esp32")) {
    Serial.println("MDNS responder started"); // start the mDNS responder and print a message on success
  }
}

void loop() {
  // reconnect to WiFi if connection was lost
  connectWiFi();
  
  // the keyboard functions...
  if(check_3x4keyboard_clicked()) {
    // show that now the code can be submitted
    digitalWrite(readyForInputLED, HIGH);
    
    // read the code and make checks, to open or close the safe
    String output = readCodeFrom3x4keyboard(true);
    if (output == getCode() && safeOpened == false && output != "newCodeFinished") {
      // open the safe
      Serial.println("Opening the safe...");
      open();
      Serial.println("Opened the safe.");
    } else if (safeOpened == true && output != "newCodeFinished") {
      // close the safe
      Serial.println("Closing the safe...");
      close();
      Serial.println("Closed the safe.");
    } else if(output != "newCodeFinished") {  // if the ouput was not the real code AND no new code was set, THEN show that the code was wrong
      blinkLED(wrongCodeLED, 700);

      // debug message
      Serial.printf("Wrong code submitted: %s\n", output);
    }

    // now show that everything is done and no code will be read now
    digitalWrite(readyForInputLED, LOW);
  }

  // response a client if it's requesting something
  server.handleClient();
}
